import dill
import base64
import time

from ellipsis.util.root import recurse
from ellipsis import sanitize
from ellipsis import apiManager
from ellipsis.account import getInfo



def createCluster(layers, token, nodes=None, interpreter='python3.12', requirements= [], awaitTillStarted = True):

    layers = sanitize.validDictArray('layers', layers, True)
    token = sanitize.validString('token', token, True)
    nodes = sanitize.validInt('nodes', nodes, False)
    interpreter = sanitize.validString('interpreter', interpreter, True)
    requirements = sanitize.validStringArray('requirements', requirements, False)

    if type(nodes) == type(None):
        info = getInfo(token=token)
        nodes = info['plan']['maxComputeNodes']
        if nodes == 0:
            raise ValueError('You have no compute nodes in your plan. Please update your subscription')

    requirements = "\n".join(requirements)

    body = {'layers':layers, 'interpreter':interpreter, 'nodes':nodes, 'requirements':requirements}
    r = apiManager.post('/compute', body, token)

    clusterId = r['id']
    while awaitTillStarted:
        res = listClusters(token=token)['result']
        r = [x for x in res if x['id'] == clusterId][0]

        if r['status'] == 'available':
            break
        time.sleep(1)

    return {'id':clusterId}


def execute(clusterId, f, token, awaitTillCompleted=True):
    clusterId = sanitize.validUuid('clusterId', clusterId, True)
    token = sanitize.validString('token', token, True)

    if str(type(f)) != "<class 'function'>":
        raise ValueError('parameter f must be a function')

    f_bytes = dill.dumps(f)
    f_string = base64.b64encode(f_bytes)

    body = { 'file':f_string}
    apiManager.post('/compute/' + clusterId + '/execute', body, token)

    while awaitTillCompleted:
        res = listClusters(token=token)['result']
        r = [x for x in res if x['id'] == clusterId][0]
        if r['status'] == 'completed':
            break
        time.sleep(1)

    return r['result']

def parseResults(r):
    results = []
    for x in r:
        x = base64.b64decode(x)
        x = dill.loads(x)
        results = results + x

    return results

def terminateCluster(clusterId, token, awaitTillTerminated = True):
    clusterId = sanitize.validUuid('clusterId', clusterId, True)
    token = sanitize.validString('token', token, True)


    r = apiManager.post('/compute/' + clusterId + '/terminate', {}, token)

    while awaitTillTerminated:
        res = listClusters(token=token)['result']
        z = [x for x in res if x['id'] == clusterId][0]
        if z['status'] == 'stopped':
            break
        time.sleep(1)

    return r

def getClusterInfo(clusterId, token):
    res = listClusters(token=token)['result']
    r = [x for x in res if x['id'] == clusterId]
    if len(r) ==0:
        raise ValueError('No cluster found for given id')
    return r[0]

def listClusters(token, pageStart = None, listAll = True):
    token = sanitize.validString('token', token, True)


    body = { 'pageStart':pageStart }


    def f(body):
        return apiManager.get('/compute', body, token)

    r = recurse(f, body, listAll)
    for i in range(len(r['result'])):
        if 'result' in r['result'][i]:
            r['result'][i]['result'] = parseResults(r['result'][i]['result'])
    return r




from ellipsis import apiManager
from ellipsis import sanitize
from ellipsis.util.root import recurse
from ellipsis.util.root import stringToDate
from PIL import Image
from io import BytesIO
import base64


def get(pathId, timestampId, featureIds = None, userId = None, messageIds = None, listAll = True, deleted = False, extent = None, pageStart = None, token = None ):
    pathId = sanitize.validUuid('pathId', pathId, True) 
    timestampId = sanitize.validUuid('timestampId', timestampId, True) 
    token = sanitize.validString('token', token, False)
    userId = sanitize.validUuid('userId', userId, False)
    messageIds = sanitize.validUuidArray('messageIds', messageIds, False)
    featureIds = sanitize.validUuidArray('featureIds', featureIds, False)
    listAll = sanitize.validBool('listAll', listAll, True)
    deleted = sanitize.validBool('deleted', deleted, True)
    extent = sanitize.validBounds('extent', extent, False)
    pageStart = sanitize.validUuid('pageStart', pageStart, False)

    body = {'userId': userId, 'messageIds':messageIds, 'deleted':deleted, 'extent':extent, 'featureIds':featureIds, 'pageStart':pageStart}
    
    def f(body):
        r = apiManager.get('/path/' + pathId + '/vector/timestamp/' + timestampId +  '/feature/message', body, token )
        return r
    
    r = recurse(f, body, listAll)
    
    r['result'] = [{**x, 'date':stringToDate(x['date']) } for x in r['result'] ]
    
    return r

def getImage(pathId, timestampId, messageId, token = None ):
    pathId = sanitize.validUuid('pathId', pathId, True) 
    timestampId = sanitize.validUuid('timestampId', timestampId, True) 
    messageId = sanitize.validUuid('messageId', messageId, True)     
    token = sanitize.validString('token', token, False)

    r = apiManager.get('/path/' + pathId + '/vector/timestamp/' + timestampId + '/feature/message/' + messageId + '/image', None, token, False)

    if r.status_code != 200:
        raise ValueError(r.text)
        
    im = Image.open(BytesIO(r.content))
 
    return(im)


def add(pathId, timestampId, featureId, token, text = None, image=None):
    pathId = sanitize.validUuid('pathId', pathId, True) 
    timestampId = sanitize.validUuid('timestampId', timestampId, True) 
    featureId = sanitize.validUuid('featureId', featureId, True) 
    token = sanitize.validString('token', token, True)
    text = sanitize.validString('text', text, False)
    image = sanitize.validImage('image', image, False)    

    if type(image) != type(None):
        image = Image.fromarray(image.astype('uint8'))
        buffered = BytesIO()
        image.save(buffered, format="JPEG")
        img_str = str(base64.b64encode(buffered.getvalue()))
        img_str = 'data:image/jpeg;base64,' + img_str[2:-1]
    else:
        img_str = None


    body = {'image':img_str, 'text':text}
    r = apiManager.post('/path/' + pathId + '/vector/timestamp/' + timestampId + '/feature/' + featureId + '/message', body, token)
    return r


def trash(pathId, timestampId, messageId, token):
    pathId = sanitize.validUuid('pathId', pathId, True) 
    timestampId = sanitize.validUuid('timestampId', timestampId, True) 
    messageId = sanitize.validUuid('messageId', messageId, True) 
    token = sanitize.validString('token', token, True)
    body = {'trashed': True}
    r = apiManager.put('/path/' + pathId + '/vector/timestamp/' + timestampId + '/feature/message/' + messageId + '/trashed', body, token)
    return r


def recover(pathId, timestampId, messageId, token):
    pathId = sanitize.validUuid('pathId', pathId, True) 
    timestampId = sanitize.validUuid('timestampId', timestampId, True) 
    messageId = sanitize.validUuid('messageId', messageId, True) 
    token = sanitize.validString('token', token, True)
    body = {'trashed': False}
    r = apiManager.put('/path/' + pathId + '/vector/timestamp/' + timestampId + '/feature/message/' + messageId + '/trashed', body, token)
    return r





from ellipsis import apiManager
from ellipsis import sanitize


def get(pathId, token=None):
    token = sanitize.validString('token', token, False)
    pathId = sanitize.validUuid('pathId', pathId, True)

    r = apiManager.get('/path/' + pathId + '/bookmark', {}, token)
        
    return r



def add(name, bookmark, token, parentId = None, publicAccess = None, metadata=None):
    token = sanitize.validString('token', token, True)
    bookmark = sanitize.validObject('bookmark', bookmark, True)
    name = sanitize.validString('pathId', name, False)
    metadata = sanitize.validObject('metadata', metadata, False)
    publicAccess = sanitize.validObject('publicAccess', publicAccess, False)
    parentId = sanitize.validUuid('parentId', parentId, False)

    r = apiManager.post('/path/bookmark', {'name':name, 'bookmark':bookmark , 'parentId':parentId, 'publicAccess':publicAccess, 'metadata':metadata}, token)
        
    return r


def edit(pathId, token, layers=None, dems=None):
    layers = sanitize.validObject('layers', layers, False)
    dems = sanitize.validObject('dems', dems, False)
    pathId = sanitize.validUuid('pathId', pathId, True)

    r = apiManager.patch('/path/' + pathId + '/bookmark', {'layers':layers, 'dems':dems}, token)
        
    return r







from ellipsis.path.vector.root  import editFilter
from ellipsis.path.vector.root  import add
from ellipsis.path.vector.root  import editRendering


from ellipsis.path.vector import timestamp
from ellipsis.path.vector import style
from ellipsis.path.vector import featureProperty





from ellipsis import apiManager
from ellipsis import sanitize
from ellipsis.util import chunks
from ellipsis.util import loadingBar
from ellipsis.util.root import stringToDate
from ellipsis.util.root import recurse

import numpy as np
import json
import geopandas as gpd



def add(pathId, timestampId, features, token, showProgress = True, cores = 1):
    pathId = sanitize.validUuid('pathId', pathId, True)
    timestampId = sanitize.validUuid('timestampId', timestampId, True)
    token = sanitize.validString('token', token, True)
    showProgress = sanitize.validBool('showProgress', showProgress, True)
    cores = sanitize.validInt('cores', cores, True)
    features = sanitize.validGeopandas('features', features, True, cores = cores)

    features_json = features.to_json(na='drop')
    features_json = json.loads(features_json)
    features_json = np.array(features_json['features'])

    #check if first time
    firstTime = apiManager.get('/path/' + pathId, None, token)
    if not 'vector' in firstTime.keys():
        raise ValueError('Can only add features if path is of type vector')
    firstTime = len(firstTime['vector']['properties']) ==0

    if firstTime:
        print('no properties known for this timestamp adding them automatically')
        columns = features.columns
        columns = [c for c in columns if c != 'geometry']
        for c in columns:
            if 'int' in str(features.dtypes[c]) or 'Int' in str(features.dtypes[c]):
                propertyType = 'integer'
            elif 'float' in str(features.dtypes[c]) or 'Float' in str(features.dtypes[c]):
                propertyType = 'float'
            elif 'bool' in str(features.dtypes[c]):
                propertyType = 'boolean'
            elif 'datetime' in str(features.dtypes[c]):
                propertyType = 'datetime'
            else:
                propertyType = 'string'


            ###date
            body = {'name': c , 'type': propertyType , 'required': False, 'private': False}
            apiManager.post('/path/' + pathId + '/vector/property', body, token)
    indices = chunks(np.arange(features.shape[0]))


    addedIds = []
    for i in np.arange(len(indices)):
        indices_sub = indices[i]
        features_sub = features_json[indices_sub]


        featuresBody = [{'feature': features_sub[i] } for i in np.arange(len(indices_sub))]
        body = {"features":featuresBody}
        r = apiManager.post('/path/' + pathId + '/vector/timestamp/' + timestampId + '/feature', body, token)
        addedIds = addedIds + r
        if showProgress:
            loadingBar(i*3000 + len(indices_sub),features.shape[0])
        i = i+1

    return(addedIds)



def edit(pathId, timestampId, featureIds, token, features, showProgress = True, cores = 1):
    pathId = sanitize.validUuid('pathId', pathId, True)
    timestampId = sanitize.validUuid('timestampId', timestampId, True)
    token = sanitize.validString('token', token, True)
    featureIds = sanitize.validUuidArray('featureIds', featureIds, True)
    showProgress = sanitize.validBool('showProgress', showProgress, True)
    cores = sanitize.validInt('cores', cores, True)
    features = sanitize.validGeopandas('features', features, False, cores=cores)



    if type(features) != type(None) and features.shape[0] != len(featureIds):
        raise ValueError('featureIds must be of same length as the features geopandas dataframe')


    indices = chunks(np.arange(len(featureIds)),1000)
    i=0
    for i in np.arange(len(indices)):
        indices_sub = indices[i]
        featureIds_sub = list( np.array(featureIds)[indices_sub])

        if type(features) != type(None):
            features_sub = features.iloc[indices_sub]
            features_sub =features_sub.to_json(na='drop')
            features_sub = json.loads(features_sub)


        changes = [{'featureId':x[0] , 'newProperties':x[1]['properties'], 'newGeometry':x[1]['geometry']} for x in zip(featureIds_sub, features_sub['features'])]

        body = {'changes':changes}
        r = apiManager.patch('/path/' + pathId + '/vector/timestamp/' + timestampId + '/feature', body, token)

        if len(indices) > 1 and showProgress:
            loadingBar(i*1000 + len(indices_sub),len(featureIds))

    return r


def trash(pathId, timestampId, featureIds, token, showProgress = True):
    pathId = sanitize.validUuid('pathId', pathId, True)
    timestampId = sanitize.validUuid('timestampId', timestampId, True)
    token = sanitize.validString('token', token, True)
    featureIds = sanitize.validUuidArray('featureIds', featureIds, True)
    showProgress = sanitize.validBool('showProgress', showProgress, True)

    indices = chunks(np.arange(len(featureIds)),1000)
    i=0
    for i in np.arange(len(indices)):
        indices_sub = indices[i]
        featureIds_sub = list(np.array(featureIds)[indices_sub])

        body = {'featureIds': featureIds_sub, 'trashed': True}
        r = apiManager.put('/path/' + pathId + '/vector/timestamp/' + timestampId + '/feature/trashed', body, token)

        if len(indices) > 1 and showProgress:
            loadingBar(i*1000 + len(indices_sub),len(featureIds))

    return r


def recover(pathId, timestampId, featureIds, token, showProgress = True):
    pathId = sanitize.validUuid('pathId', pathId, True)
    timestampId = sanitize.validUuid('timestampId', timestampId, True)
    token = sanitize.validString('token', token, True)
    featureIds = sanitize.validUuidArray('featureIds', featureIds, True)
    showProgress = sanitize.validBool('showProgress', showProgress, True)

    indices = chunks(np.arange(len(featureIds)),1000)
    i=0
    for i in np.arange(len(indices)):
        indices_sub = indices[i]
        featureIds_sub = list(np.array(featureIds)[indices_sub])

        body = {'featureIds': featureIds_sub, 'trashed': False}
        r = apiManager.put('/path/' + pathId + '/vector/timestamp/' + timestampId + '/feature/trashed', body, token)

        if len(indices) > 1 and showProgress:
            loadingBar(i*1000 + len(indices_sub),len(featureIds))

    return r



def versions(pathId, timestampId, featureId, token = None, pageStart = None, listAll = True):
    pathId = sanitize.validUuid('pathId', pathId, True)
    timestampId = sanitize.validUuid('timestampId', timestampId, True)
    token = sanitize.validString('token', token, False)
    featureId = sanitize.validUuid('featureId', featureId, True)
    pageStart = sanitize.validUuid('pageStart', pageStart, False)
    listAll = sanitize.validBool('listAll', listAll, True)

    body = {'returnType':'all'}
    def f(body):
        return apiManager.get('/path/' + pathId + '/vector/timestamp/' + timestampId + '/feature/' + featureId + '/version', body, token)

    r = recurse(f, body, listAll)

    features = [ x['feature'] for x in r['result']]
    dates = [stringToDate(x['date']) for x in r['result'] ]
    usernames = [x['user']['username'] for x in r['result'] ]
    userIds = [x['user']['id'] for x in r['result'] ]

    sh = gpd.GeoDataFrame.from_features(features)
    sh['username'] = usernames
    sh['userId'] = userIds
    sh['dates'] = dates

    sh.crs = {'init': 'epsg:4326'}
    r['result'] = sh
    return(r)





from ellipsis import apiManager
from ellipsis import sanitize
from ellipsis.util.root import recurse
import requests

def logIn(username, password, validFor = None):

        username = sanitize.validString('username', username, True)
        password = sanitize.validString('password', password, True)
        validFor = sanitize.validInt('validFor', validFor, False)

        json = {'username': username, 'password': password, 'validFor': validFor}

        r = apiManager.call(requests.post,'/account/login', body=json, token=None, crash=False)
        if r.status_code == 400:
            x = r.json()
            if x['message'] == "No password configured.":
                raise ValueError("You cannot login with your Google credentials in the Python module. You need to configure an Ellipsis Drive specific password. You can do this on https://app.ellipsis-drive.com/account-settings/security")
        if r.status_code != 200:
            raise ValueError(r.text)

        r = r.json()
        token = r['token']

        return(token)


def getInfo(token):
    token = sanitize.validString('token', token, True)

    r = apiManager.get( '/account', body={}, token=token)


    return r

def listRoot(rootName, token, pathTypes= None, pageStart = None, listAll = True):
    token = sanitize.validString('token', token, True)
    rootName = sanitize.validString('rootName', rootName, True)
    pageStart = sanitize.validUuid('pageStart', pageStart, False)
    listAll = sanitize.validBool('listAll', listAll, True)        
    pathTypes = sanitize.validObject('pathTypes', pathTypes, False)
    if type(pathTypes) == type(None):
        pathTypes = ['folder', 'raster', 'vector', 'file']
        

    url = "/account/root/" + rootName
    body = {"type": pathTypes, "pageStart": pageStart}


    def f(body):
        r = apiManager.get(url, body, token)
        return r

    r = recurse(f, body, listAll)
        
    return r




